import 'dart:ffi';
import 'package:meta/meta.dart';
import 'array.dart';
import 'slice.dart';

/// Instead of doing many little allocations, [Arena]
/// allows you to make one large allocation which can then be sliced.
/// This can improve performance and simplify memory management.
/// ````
/// final buffer = Uint8Array.allocate(10);
/// final arena = Arena(buffer, buffer.ptr.size);
/// final array = arena.newArray(5);
/// final array2 = arena.newArray(5);
/// void some_native_c_api(array.ptr, array.length, array2.ptr, array2.length);
/// arena.free();
/// ````
@sealed
class Arena<T extends NativeType> {
  int _index = 0;

  final Array<T> _buffer;
  final int _elementSize;

  /// Left free space of Arena
  int get freeSpaceLeft => _buffer.length - _index;

  /// Allocates a buffer of [length]
  Arena(this._buffer, {required int elementSize}) : _elementSize = elementSize;

  /// Returns a view into [_buffer].
  /// When [length] exceeds [freeSpaceLeft] it throws a [RangeError].
  Slice<T> slice(int length) {
    assert(length > 0);
    if (freeSpaceLeft < length) {
      throw RangeError('Not enough free space');
    }
    final arr = Slice.fromArray(_buffer, length,
        elementSizeInBytes: _elementSize, startIndex: _index);
    _index += length;
    return arr;
  }

  void free() {
    _buffer.free();
  }
}
